/*  file.c  (C) 1997 Mark A. Livingston */
/* Any non-commercial use of this code is permitted so long as */
/* the above copyright line appears in the file.  Please contact */
/* the author regarding permission for commercial use. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <quat.h>
#include "quatext.h"

#define SQR(a) ( (a) * (a) )
#define SQRT1_3 0.5773502691896257645

static FILE *fp = NULL;
static FILE *bfp = NULL;

int
OpenFile( char *fname )
{
    fp = fopen( fname, "r" );
    if( fp == NULL )
      return errno;
    else {
      rewind( fp );
      return 0;
    }
}

int
CloseFile( void )
{
    if( fclose( fp ) == EOF )
      return EOF;
    else
      return 0;
}

void
NextSample( q_xyz_quat_type *v )
{
    char buf[256], first[80];

    do {
      fgets( buf, 255, fp );
      sscanf( buf, "%s", first );
    } while( !strncmp( first, "/*"/**/, 2 ) );

    /* Extract into position/quaternion */
    sscanf( buf, "{ { %lf, %lf, %lf }, { %lf, %lf, %lf, %lf } },",
	    &v->xyz[ Q_X ], &v->xyz[ Q_Y ], &v->xyz[ Q_Z ],
	    &v->quat[ Q_X ], &v->quat[ Q_Y ], &v->quat[ Q_Z ], &v->quat[Q_W] );

#if 0
    if( v->quat[ Q_W ] < 0.0 ) {
      v->quat[ Q_X ] = -v->quat[ Q_X ];
      v->quat[ Q_Y ] = -v->quat[ Q_Y ];
      v->quat[ Q_Z ] = -v->quat[ Q_Z ];
      v->quat[ Q_W ] = -v->quat[ Q_W ];
    }
#endif

    return;
}

/* A hack to enable ASCII-to-binary conversion, by getting times explicitly */
void
NextTime( double *time )
{
    char buf[256];
    fgets( buf, 255, fp );
    sscanf( buf, "    /* time for samples = %20.17lf seconds */", time );
    return;
}

static q_vec_type V = { SQRT1_3, SQRT1_3, SQRT1_3 };

void
NextPoint( q_xyz_quat_type *v, double *diag, double *angle, int nSamp )
{
    int i;
    q_xyz_quat_type *s;
    q_vec_type minPos, maxPos, tmp;
    double *weights;

    /* Alloc, not error checking */
    s = ( q_xyz_quat_type *) malloc( nSamp * sizeof( q_xyz_quat_type ) );
    weights = (double *) malloc( nSamp * sizeof( double ) );

    /* Read nSamp samples */
    for( i = 0; i < nSamp; i++ ) {
      NextSample( &s[i] );
      weights[i] = 1.0 / (double) nSamp;
    }
    /* compute the average */
    qx_xyz_quat_wgt_avg( v, s, weights, nSamp );

    /* compute the diagonal of the bounding box */
    q_vec_copy( minPos, s[0].xyz );
    q_vec_copy( maxPos, s[0].xyz );
    for( i = 1; i < nSamp; i++ ) {
      if( s[i].xyz[Q_X] < minPos[Q_X] ) minPos[Q_X] = s[i].xyz[Q_X];
      else if( s[i].xyz[Q_X] > maxPos[Q_X] ) maxPos[Q_X] = s[i].xyz[Q_X];
      if( s[i].xyz[Q_Y] < minPos[Q_Y] ) minPos[Q_Y] = s[i].xyz[Q_Y];
      else if( s[i].xyz[Q_Y] > maxPos[Q_Y] ) maxPos[Q_Y] = s[i].xyz[Q_Y];
      if( s[i].xyz[Q_Z] < minPos[Q_Z] ) minPos[Q_Z] = s[i].xyz[Q_Z];
      else if( s[i].xyz[Q_Z] > maxPos[Q_Z] ) maxPos[Q_Z] = s[i].xyz[Q_Z];
    }
    *diag = q_vec_distance( minPos, maxPos );

    /*
    ** Rotate the vector \sqrt{1/3} x ( 1, 1, 1 ) by the quaternion and
    ** compute the size of the bounding box of the resultant endpoints.
    */
    q_xform( minPos, s[0].quat, V );
    q_vec_copy( maxPos, minPos );
    for( i = 1; i < nSamp; i++ ) {
      q_xform( tmp, s[i].quat, V );
      if( tmp[ Q_X ] < minPos[ Q_X ] ) minPos[ Q_X ] = tmp[ Q_X ];
      else if( tmp[ Q_X ] > maxPos[ Q_X ] ) maxPos[ Q_X ] = tmp[ Q_X ];
      if( tmp[ Q_Y ] < minPos[ Q_Y ] ) minPos[ Q_Y ] = tmp[ Q_Y ];
      else if( tmp[ Q_Y ] > maxPos[ Q_Y ] ) maxPos[ Q_Y ] = tmp[ Q_Y ];
      if( tmp[ Q_Z ] < minPos[ Q_Z ] ) minPos[ Q_Z ] = tmp[ Q_Z ];
      else if( tmp[ Q_Z ] > maxPos[ Q_Z ] ) maxPos[ Q_Z ] = tmp[ Q_Z ];
    }
    *angle = q_vec_distance( minPos, maxPos );

    free( s );

    return;
}

/* Reads an 4 x 4 matrix from the open file, fills in matrix m */
void
NextMatrix( q_matrix_type m )
{
    int i, j;
    char buf[256], first[80], *cur;

    for( i = 0; i < 4; i++ ) {
      /* Find the next row of the matrix */
      do {
	fgets( buf, 255, fp );
	sscanf( buf, "%s", first );
      } while( !strncmp( first, "/*"/**/, 2 ) );

      /* Extract n columns */
      for( cur = &buf[0], j = 0; j < 4; j++, cur += strlen( first ) + 1 ) {
	sscanf( cur, "%s", first );
	sscanf( first, "%lf", &m[i][j] );
      } /* end for(j,cur) */
    } /* end for(i) */

    return;
}

int
OpenBFile( char *fname )
{
    bfp = fopen( fname, "rb" ); /* b is "inconsequential" says the man page */
    if( bfp == NULL )
      return errno;
    else {
      rewind( bfp );
      return 0;
    }
}

int
CloseBFile( void )
{
    if( fclose( bfp ) == EOF )
      return EOF;
    else
      return 0;
}

int
NextBSample( q_xyz_quat_type *v )
{
    /*
    ** The binary file implementation assumes that the user has left the
    ** pointer in the file in the right place, so this is just a call to
    ** fread( ).  The implementation further assumes the type and size of
    ** the q_xyz_quat_type.
    */
    return fread( v, sizeof( double ), 7, bfp );
}

void
NextBPoint( q_xyz_quat_type *v, double *diag, double *angle, int nSamp )
{
    int i;
    q_xyz_quat_type *s;
    q_vec_type minPos, maxPos, tmp;
    double *weights;

    /* Alloc, not error checking */
    s = ( q_xyz_quat_type * ) malloc( nSamp * sizeof( q_xyz_quat_type ) );
    weights = ( double * ) malloc( nSamp * sizeof( double ) );

    /* read nSamp samples */
    for( i = 0; i < nSamp; i++ ) {
      (void) NextBSample( &s[i] );
      weights[i] = 1.0 / (double) nSamp;
    }
    /* compute the average */
    qx_xyz_quat_wgt_avg( v, s, weights, nSamp );

    /* compute the diagonal of the bounding box */
    q_vec_copy( minPos, s[0].xyz );
    q_vec_copy( maxPos, s[0].xyz );
    for( i = 1; i < nSamp; i++ ) {
      if( s[i].xyz[Q_X] < minPos[Q_X] ) minPos[Q_X] = s[i].xyz[Q_X];
      else if( s[i].xyz[Q_X] > maxPos[Q_X] ) maxPos[Q_X] = s[i].xyz[Q_X];
      if( s[i].xyz[Q_Y] < minPos[Q_Y] ) minPos[Q_Y] = s[i].xyz[Q_Y];
      else if( s[i].xyz[Q_Y] > maxPos[Q_Y] ) maxPos[Q_Y] = s[i].xyz[Q_Y];
      if( s[i].xyz[Q_Z] < minPos[Q_Z] ) minPos[Q_Z] = s[i].xyz[Q_Z];
      else if( s[i].xyz[Q_Z] > maxPos[Q_Z] ) maxPos[Q_Z] = s[i].xyz[Q_Z];
    }
    *diag = q_vec_distance( minPos, maxPos );

    /*
    ** Rotate the vector \sqrt{1/3} x ( 1, 1, 1 ) by the quaternion and
    ** compute the size of the bounding box of the resultant endpoints.
    */
    q_xform( minPos, s[0].quat, V );
    q_vec_copy( maxPos, minPos );
    for( i = 1; i < nSamp; i++ ) {
      q_xform( tmp, s[i].quat, V );
      if( tmp[ Q_X ] < minPos[ Q_X ] ) minPos[ Q_X ] = tmp[ Q_X ];
      else if( tmp[ Q_X ] > maxPos[ Q_X ] ) maxPos[ Q_X ] = tmp[ Q_X ];
      if( tmp[ Q_Y ] < minPos[ Q_Y ] ) minPos[ Q_Y ] = tmp[ Q_Y ];
      else if( tmp[ Q_Y ] > maxPos[ Q_Y ] ) maxPos[ Q_Y ] = tmp[ Q_Y ];
      if( tmp[ Q_Z ] < minPos[ Q_Z ] ) minPos[ Q_Z ] = tmp[ Q_Z ];
      else if( tmp[ Q_Z ] > maxPos[ Q_Z ] ) maxPos[ Q_Z ] = tmp[ Q_Z ];
    }
    *angle = q_vec_distance( minPos, maxPos );

    free( s );

    return;
}

/* We do not need a binary read of a matrix, since those are always ASCII. */

/*
** This function is new to the binary file interface, since we must account
** for the time values embedded in the file.
*/
void
NextBPointPair( q_xyz_quat_type *v1, double *diag1, double *angle1, int nSamp1,
                q_xyz_quat_type *v2, double *diag2, double *angle2, int nSamp2,
                double *time )
{
    NextBPoint( v1, diag1, angle1, nSamp1 );
    NextBPoint( v2, diag2, angle2, nSamp2 );
    fread( time, sizeof( double ), 1, bfp );
    return;
}

/*
** All of these functions are new to the binary interface, since we wrote with
** printf( ) family functions before.  Note that the FILE pointer is the same
** as for reading binary files, thus you can only use one at a time.
*/
int
OpenBFileW( char *fname )
{
    bfp = fopen( fname, "wb" ); /* w = truncate or create, b = nothing */
    if( bfp == NULL )
      return errno;
    else
      return 0;
}

/* not using a WriteSample - this is just as easy */
void
WritePoint( q_xyz_quat_type *s, int nSamp )
{
    int i;

    for( i = 0; i < nSamp; i++ )
      fwrite( &s[i], sizeof( double ), 7, bfp );

    return;
}

void
WriteTime( double *time )
{
    fwrite( time, sizeof( double ), 1, bfp );
    return;
}

void
WritePointPair( q_xyz_quat_type *s1, int nSamp1,
                q_xyz_quat_type *s2, int nSamp2, double *time )
{
    WritePoint( s1, nSamp1 );
    WritePoint( s2, nSamp2 );
    fwrite( time, sizeof( double ), 1, bfp );
    return;
}
